(*******************************************************************************)
(*  © Université de Lille, The Pip Development Team (2015-2024)                *)
(*                                                                             *)
(*  This software is a computer program whose purpose is to run a minimal,     *)
(*  hypervisor relying on proven properties such as memory isolation.          *)
(*                                                                             *)
(*  This software is governed by the CeCILL license under French law and       *)
(*  abiding by the rules of distribution of free software.  You can  use,      *)
(*  modify and/ or redistribute the software under the terms of the CeCILL     *)
(*  license as circulated by CEA, CNRS and INRIA at the following URL          *)
(*  "http://www.cecill.info".                                                  *)
(*                                                                             *)
(*  As a counterpart to the access to the source code and  rights to copy,     *)
(*  modify and redistribute granted by the license, users are provided only    *)
(*  with a limited warranty  and the software's author,  the holder of the     *)
(*  economic rights,  and the successive licensors  have only  limited         *)
(*  liability.                                                                 *)
(*                                                                             *)
(*  In this respect, the user's attention is drawn to the risks associated     *)
(*  with loading,  using,  modifying and/or developing or reproducing the      *)
(*  software by the user in light of its specific status of free software,     *)
(*  that may mean  that it is complicated to manipulate,  and  that  also      *)
(*  therefore means  that it is reserved for developers  and  experienced      *)
(*  professionals having in-depth computer knowledge. Users are therefore      *)
(*  encouraged to load and test the software's suitability as regards their    *)
(*  requirements in conditions enabling the security of their systems and/or   *)
(*  data to be ensured and,  more generally, to use and operate it in the      *)
(*  same conditions as regards security.                                       *)
(*                                                                             *)
(*  The fact that you are presently reading this means that you have had       *)
(*  knowledge of the CeCILL license and that you accept its terms.             *)
(*******************************************************************************)

(** * Summary 
    This file contains the invariant of [getTableAddr] some associated lemmas *)
Require Import Pip.Model.ADT Pip.Model.Hardware Pip.Model.Lib Pip.Model.MAL.
Require Import Pip.Core.Internal.

Require Import Pip.Proof.Consistency Pip.Proof.DependentTypeLemmas Pip.Proof.InternalLemmas
               Pip.Proof.Isolation Pip.Proof.StateLib Pip.Proof.WeakestPreconditions.

Require Import Invariants.

Require Import List Coq.Logic.ProofIrrelevance Lia Compare_dec Lt EqNat.

Lemma getTableAddr  (indirection : page) (va : vaddr) (l : level) P currentPart idxroot: 
{{fun s => P s /\ consistency s /\ In currentPart (getPartitions pageRootPartition s) /\
            (idxroot = idxPageDir \/ idxroot = idxShadow1 \/ idxroot = idxShadow2) /\
            ( exists (tableroot : page),
                nextEntryIsPP currentPart idxroot tableroot s /\
                tableroot <> pageDefault /\
                ( (tableroot = indirection /\ Some l = StateLib.getNbLevel ) \/
                (exists nbL stop, Some nbL = StateLib.getNbLevel /\ stop <= nbL /\
                StateLib.getIndirection tableroot va nbL stop s = Some indirection /\
                indirection <> pageDefault  /\
                l = CLevel (nbL - stop)))) }}

getTableAddr indirection va l

{{fun (table : page) (s : state) => P s  /\  
                 (getTableAddrRoot' table idxroot currentPart va s /\ table = pageDefault \/ 
                 (getTableAddrRoot table idxroot currentPart va s /\  table<> pageDefault  /\ 
                    ( forall idx,  StateLib.getIndexOfAddr va levelMin = idx -> 
                    ( (isVE table idx s /\ idxroot = idxShadow1) \/ 
                      (isVA table idx s /\ idxroot = idxShadow2) \/
                      (isPE table idx s /\ idxroot = idxPageDir) )  ) ) )
 }}.
Proof.
unfold Internal.getTableAddr.
assert(Hsize : nbLevel   <= nbLevel) by lia.
assert (Hlevel : l < nbLevel).
destruct l. simpl;lia.
revert Hsize Hlevel.
revert indirection l.
generalize nbLevel at 1 3 4.
induction n; simpl.
+ intros. destruct l.
  simpl in *.
  lia.  
+ intros.
  eapply WP.bindRev.
  eapply WP.weaken.
  eapply Invariants.Level.eqb.
  intros.
  simpl.
  pattern s in H.
  eapply H.
  intros isFstLevel.    
  case_eq isFstLevel. 
   - intros Hfstlevel. eapply WP.weaken.
     eapply WP.ret .
     intros s Hinv. simpl.
     split. intuition.
     unfold consistency in Hinv.    
     clear IHn.
     destruct Hinv as ((Hp & (Hprp & HPD & Hsh1 & Hsh2& Hcprpl & HmappedNodup) & Hcp & Hroot & rootind & 
      Hcurpd & Hpdnotnull  & [Hindirection | Hindirection]) & H).
     * (** tableroot = indirection **)
       destruct Hindirection.
       right.
       split. 
       unfold getTableAddrRoot.
       symmetry in H. clear Hfstlevel.
       rename H into Hfstlevel.
       unfold getTableAddrRoot.
       split.
       assumption.
       intros. exists l. 
        split. assumption.
        exists (l+1). 
        split. reflexivity.
       apply levelEqBEqNatTrue in Hfstlevel.
       rewrite Hfstlevel.
       unfold levelMin.
       unfold CLevel. simpl. 
       case_eq (lt_dec 0 nbLevel).
       intros. simpl. subst.
        unfold nextEntryIsPP in H, Hcurpd.
        destruct (StateLib.Index.succ idxroot); [| now contradict H1].
        unfold levelEq. 
       assert (Nat.eqb ({| l := 0; Hl := ADT.CLevel_obligation_1 0 l0 |}) levelMin = true).
       apply NPeano.Nat.eqb_eq. simpl. 
       unfold levelMin. unfold CLevel. rewrite H2. simpl. trivial.
       rewrite H0.
       destruct(lookup currentPart i (memory s) pageEq idxEq);
       [| now contradict H1]. 
       destruct v; try now contradict Hcurpd.
       subst ; trivial. 
       intros.
       unfold levelEq.
       assert (0 < nbLevel). apply nbLevelNotZero.
       intros. subst.  now contradict H2.
       split.
       subst. trivial.
       intros.     
       apply  fstIndirectionContainsValue_nbLevel_1 with va l currentPart;trivial.
       destruct Hroot as [Hrootpd | [Hrootsh1 |Hrootsh2]];subst;trivial. 
       unfold currentPartitionInPartitionsList in *.
       subst. symmetry. assumption.
       subst. assumption. subst. assumption.   
     * (** tableroot <> indirection *)  
      right. split. 
       destruct Hindirection as  (nbL & stop & Hl & Hind & Hnotnull& Hstop & Hstople).
       unfold getTableAddrRoot.
       subst.
       split. assumption.
       intros. 
       exists nbL.
        split. assumption.
       exists (nbL+1). split.
       reflexivity.
       rename H into Hfstlevel.
       symmetry in Hfstlevel.
       apply levelEqBEqNatTrue in Hfstlevel.
       unfold levelMin in Hfstlevel.
       apply CLevelMinusEq0 in Hfstlevel.
       apply le_lt_or_eq in Hind.
       destruct nbL.
       simpl in *.
       unfold nextEntryIsPP in H0, Hcurpd.
        destruct (StateLib.Index.succ idxroot); [| now contradict H0]. 
       destruct (lookup currentPart i (memory s) pageEq idxEq);
       [| now contradict H0]. 
       destruct v; try now contradict Hcurpd.
       subst ; trivial.
       destruct Hind.
       destruct Hfstlevel.
       {
       unfold CLevel in H0. 
       case_eq(lt_dec stop nbLevel);intros.
       
       rewrite H1 in H0.
       inversion H0. lia.
        lia. }
        { lia. }
        { rewrite H in Hnotnull.
       
          apply getIndirectionStopLevelGT with l.
          simpl. lia. simpl. lia. assumption. }
     
       split.
        destruct Hindirection as  (nbL & stop & Hl & Hind & Hnotnull& Hstop & Hstople).
        assumption.   
       intros idx Hidx.
       apply lastIndirectionContainsValueRec with l rootind va currentPart;trivial.
       subst. symmetry; assumption. 
       destruct Hroot as [Hrootpd | [Hrootsh1 |Hrootsh2]];subst;trivial.
      - intros Hnotfstlevel.
      eapply WP.bindRev.
      (** getIndexOfAddr **)
      eapply WP.weaken.
      eapply Invariants.getIndexOfAddr.
      intros. simpl. pattern s in H.
      eapply H.
      simpl.
      intros idxind.
      eapply WP.bindRev.
      (** readPhyEntry **)
      { eapply WP.weaken.
        eapply Invariants.readPhyEntry.
        simpl. intros.
        split.
        + pattern s in H. eapply H.
        + clear IHn.
           destruct H as ( (( Hp & (Hprp & HPD & Hsh1 & Hsh2 & Hcprpl& HmappedNodup) 
           & Hcp & Hroot & currentpd & Hcurpd & Hpdnotnull  &   [Hindirection | Hindirection]) 
            & Hlvl) & Hidx). 
(*           destruct H as ((Hp & (* (Hprp & HPD & Hsh1 & Hsh2 & Hcprpl& HmappedNodup) *) Hcons 
          & Hcp & Hroot & currentpd & Hcurpd & Hpdnotnull   (* &   [Hindirection | Hindirection]  *) ) & Hidx)).  *)
          - subst. destruct Hindirection. subst. 
            eapply fstIndirectionContainsPENbLevelGT1;try eassumption.
            symmetry; trivial.  
            destruct Hroot as [Hrootpd | [Hrootsh1 |Hrootsh2]];subst;trivial.
            reflexivity.
          - subst.  eapply middleIndirectionsContainsPE;try eassumption.
            symmetry; trivial.
            destruct Hroot as [Hroot | [Hroot | Hroot] ];subst;trivial. }
      intros nextindirection.
      simpl.
      eapply WP.bindRev. 
      (** comparePageToNull **)
      eapply WP.weaken.
      eapply Invariants.comparePageToNull.
      simpl. intros.
      pattern s in H.
      eapply H.
      intros isNull.
      case_eq isNull.
      {         intros nextindisnull.

        eapply WP.weaken.
        eapply WP.ret .
        intros.
        simpl.
        unfold getTableAddrRoot'.
        split.
        intuition.
        left. 
        assert ((exists tableroot : page,
        nextEntryIsPP currentPart idxroot tableroot s /\
        tableroot <> pageDefault /\
        (tableroot = indirection /\ Some l = StateLib.getNbLevel \/
        (exists (nbL : level) (stop : nat),
        Some nbL = StateLib.getNbLevel /\
        stop <= nbL /\
        getIndirection tableroot va nbL stop s = Some indirection /\
        indirection <> pageDefault /\ l = CLevel (nbL - stop))))) by intuition.
        destruct H0.
        destruct H0.
        destruct H1.
        split.
        split.  intuition.
        intros.  
        destruct H2.

        { destruct H2.  
        intros.
        exists l.
        split; trivial.
        exists (nbLevel -1).
        assert (false = (Nat.eqb l levelMin)) as Hfstlevel by intuition.
        symmetry in Hfstlevel.
        apply levelEqBEqNatFalse0 in Hfstlevel.
        apply and_assoc.
        split. 
        unfold StateLib.getNbLevel in H4.
        case_eq(gt_dec nbLevel 0); intros.
        rewrite H5 in H4.
        inversion H4. simpl.
        destruct l.
        inversion H7. 
        split; trivial. simpl in *. subst.   lia.
        
         assert (nbLevel > 0) by apply nbLevelNotZero.
        lia. 
        subst.
        destruct H.
        destruct H.
        destruct H.
        destruct H. 
        clear H.
        unfold nextEntryIsPP in *.
        destruct (StateLib.Index.succ idxroot); [| now contradict H0].
        destruct (lookup currentPart i (memory s) pageEq idxEq); [| now contradict H0].
        destruct v ; try now contradict H0.
        subst.
        unfold isEntryPage in *.
        assert (nbLevel > 0) by apply nbLevelNotZero.
        (*   assert ((nbLevel -1) +1 = nbLevel) by lia.
        rewrite <- H0.
        destruct (nbLevel - 1) .
        simpl.
        rewrite <- H7.
        unfold readPhyEntry.   *)
        case_eq(lookup p (StateLib.getIndexOfAddr va l) (memory s) pageEq idxEq); intros;
        rewrite H0 in H5.
        case_eq v; intros  ; rewrite H3 in H5; try now contradict H5.

        subst.
        assert (nbLevel -1 > 0). 
        symmetry in H7.
        apply levelEqBEqNatFalse0 in H7.
        unfold StateLib.getNbLevel in H4.
        case_eq  (gt_dec nbLevel 0); intros.
        rewrite H3 in H4.
        inversion H4.
        destruct l.
        simpl in *.
        inversion H6.
        subst. assumption.
        assert (nbLevel > 0) by apply nbLevelNotZero.
        lia.    
        destruct (nbLevel - 1) .
        simpl.
        now contradict H3.
        simpl.
        rewrite <- H7.
        unfold StateLib.readPhyEntry.  
        rewrite H0.    rewrite H2.
        trivial.
        now contradict H5. }
        {
        (*  simpl.
         
        simpl. 
        simpl.
        *)   
        destruct H.
        destruct H.
        destruct H.
        destruct H.

        clear H. 
        destruct H2.
        destruct H.
        destruct H.
        destruct H2.
        destruct H8.
        destruct H9.
        apply beq_nat_true in H4.
        subst.
        symmetry in H7.

        unfold nextEntryIsPP in *.   
        destruct (StateLib.Index.succ idxroot); [| now contradict H0].
        destruct (lookup currentPart i (memory s) pageEq idxEq); [| now contradict H0].
        destruct v ; try now contradict H0.
        subst.
        unfold isEntryPage in H5.
        case_eq(lookup indirection (StateLib.getIndexOfAddr va (CLevel (x0 - x1))) 
        (memory s) pageEq idxEq); intros; rewrite H0 in H5; try now contradict H5.

        case_eq v; intros; rewrite H3 in H5; try now contradict H5. 
        subst.
        assert ( levelEq (CLevel (x0 - x1)) levelMin = false). trivial.
        apply levelEqBEqNatFalse0 in H7.
        exists (CLevel (nbLevel -1)).

        split; trivial.
        unfold StateLib.getNbLevel.
        case_eq (gt_dec nbLevel 0); intros.
        unfold CLevel. 
        case_eq( lt_dec (nbLevel - 1) nbLevel ) ; intros.
        f_equal.
        f_equal.
        apply proof_irrelevance.
        lia.
        assert (nbLevel > 0) by apply nbLevelNotZero.
        lia.
        exists (nbLevel -1).
        apply and_assoc.
        split.
        unfold CLevel.
        case_eq ( lt_dec (nbLevel - 1) nbLevel ) ; intros.
        simpl. split.
        unfold   StateLib.getNbLevel in H.
        case_eq  (gt_dec nbLevel 0); intros.
        rewrite H6 in *.
        inversion H.
        destruct x0.
        simpl in *.
        inversion H11.
        subst.
        unfold CLevel in H7.
        case_eq (lt_dec (nbLevel - 1 - x1) nbLevel ); intros; rewrite H10 in *.
        simpl in H7.
        lia.
        lia.
        assert (nbLevel > 0) by apply nbLevelNotZero.
        lia. 
          lia.
          assert (nbLevel > 0) by apply nbLevelNotZero.
        lia.
        assert (x1 < x0).
        apply level_gt.
        destruct x0.
        simpl in *. lia. assumption.
        clear H2.
        
        assert (getIndirection p va x0 (x1+1) s = Some (pa p0)).

        apply  getIndirectionProp with (StateLib.getIndexOfAddr va (CLevel (x0 - x1)))indirection;
        try lia; try trivial.
        clear H8 H0.
        apply getIndirectionNbLevelEq with (x1+1).
        lia.
        unfold CLevel.
        case_eq (lt_dec (nbLevel - 1) nbLevel); intros.
        simpl; trivial.
        assert (0 < nbLevel) by apply nbLevelNotZero.
        lia.
        assert (x0 =  (CLevel (nbLevel - 1))).

        unfold StateLib.getNbLevel in *.
        case_eq (  gt_dec nbLevel 0 ); intros; rewrite H0 in *.
        inversion H.
        destruct x0.
        simpl in *.
        inversion H8.
        subst.
        unfold CLevel.
        case_eq (lt_dec (nbLevel - 1) nbLevel); intros.
        f_equal.
        apply proof_irrelevance.
        assert (0 < nbLevel) by apply nbLevelNotZero.
        lia.
        assert (0 < nbLevel) by apply nbLevelNotZero.
        lia.
        subst.
        lia.
        assert (x0 = (CLevel (nbLevel - 1))).
        unfold StateLib.getNbLevel in *.
        case_eq (  gt_dec nbLevel 0 ); intros; rewrite H0 in *.
        inversion H.
        destruct x0.
        simpl in *.
        inversion H8.
        subst.
        unfold CLevel.
        case_eq (lt_dec (nbLevel - 1) nbLevel); intros.
        f_equal.
        apply proof_irrelevance.
        assert (0 < nbLevel) by apply nbLevelNotZero.
        lia.
        assert (0 < nbLevel) by apply nbLevelNotZero.
        lia.
        rewrite <- H0.
        rewrite H2.
        f_equal.
        destruct p0.
        simpl in *. 
        unfold pageDefault in * .
        unfold CPage in *.
        case_eq (lt_dec 0 nbPage); intros.
        rewrite H6 in *.
        destruct pa.
        simpl in *.
        subst. f_equal.
        apply proof_irrelevance.
        rewrite H6 in *.
        destruct pa. destruct page_d. simpl in *.
        subst.
        f_equal.
        apply proof_irrelevance.

        }
        trivial.
      }
      { intros nextindnotnull.
        simpl.
        eapply WP.bindRev.
        eapply WP.weaken.
        eapply Invariants.Level.pred.
        intros.
        simpl.
        split.
        pattern s in H.
        eapply H.
        clear Hnotfstlevel. 
        assert ( false = levelEq l levelMin) as Hnotfstlevel by intuition.
        
        unfold levelEq in Hnotfstlevel.
        subst.
        symmetry in Hnotfstlevel. 
        apply beq_nat_false in Hnotfstlevel. 
        unfold levelMin in Hnotfstlevel.
        unfold CLevel in Hnotfstlevel.
        case_eq ( lt_dec 0 nbLevel).
        intros. subst.
        rewrite H0 in Hnotfstlevel.
        simpl in *.
        lia.
        intros.
        assert (0 < nbLevel).
        apply nbLevelNotZero.
        lia.
        intro levelpred. 
        simpl in *. 
        (** next step **)
        unfold hoareTriple in *.
        intros.
        generalize (IHn nextindirection levelpred  );clear IHn;intro IHn.
        subst.
        +  assert ( P s /\
      consistency s /\
      In currentPart (getPartitions pageRootPartition s) /\
      (idxroot = idxPageDir \/ idxroot = idxShadow1 \/ idxroot = idxShadow2) /\
      (exists tableroot : page,
         nextEntryIsPP currentPart idxroot tableroot s /\
         tableroot <> pageDefault /\
         (tableroot = nextindirection /\ Some levelpred = StateLib.getNbLevel \/
          (exists (nbL : level) (stop : nat),
             Some nbL = StateLib.getNbLevel /\
             stop <= nbL /\
             getIndirection tableroot va nbL stop s = Some nextindirection /\
             nextindirection <> pageDefault /\ levelpred = CLevel (nbL - stop)))) ).
        { clear IHn.
          destruct H as ( ((((( Hp & Hcons & Hcurpart & Hroot &  H8) & H1 ) & H2 ) & H3 ) & H4)& Hbidon).
          split;trivial. split;trivial.
          split;trivial. split;trivial. 
          destruct H8 as ( pd & idxpd & H7).
          destruct H7 as ( Hnotnull & 
                          [(Hpd & Hlvl) | (nbL & stop & Hnbl &Hstople &Hind & Hindnotnull & Hstop)]) .
          + exists pd. subst. 
            split;trivial. split;trivial. right.
            exists l, 1.
            split;trivial.
            subst.
            split.
            symmetry in H1.
            rename H1 into Hnotfstlevel. 
            apply levelEqBEqNatFalse0 in Hnotfstlevel.
            destruct l. lia.
            split.
            rename H3 into H2.  
            unfold isEntryPage in H2.
            case_eq (lookup indirection (StateLib.getIndexOfAddr va l)  (memory s) pageEq idxEq);
            [intros v H0 | intros H0];rewrite H0 in H2; try now contradict H2.
            destruct v ; try now contradict H2. subst.
             
            apply getIndirectionStop1   with (StateLib.getIndexOfAddr va l) ; try assumption.
            symmetry; assumption.
            trivial.
            split.
            unfold not.
            intros.
            apply beq_nat_false in H4.
            rewrite H in H4. 
            now contradict H4.
            apply levelPredMinus1. symmetry;trivial. trivial.
            
          + exists pd. subst. 
            split;trivial. split;trivial.
            right.
            exists nbL, (stop +1).
            split;trivial.
            split.
            rename H1 into Hnotfstlevel.
            symmetry in Hnotfstlevel.
            apply levelEqBEqNatFalse0 in Hnotfstlevel.
             
            assert(nbL - stop < nbLevel).
            destruct nbL.
            simpl in *.
            lia.
            apply level_gt in Hnotfstlevel.
            lia. assumption.
            split.  
            subst.
            unfold isEntryPage in H3.
            
            case_eq (lookup indirection (StateLib.getIndexOfAddr va (CLevel (nbL - stop)))  
            (memory s) pageEq idxEq); [intros v H0 | intros H0]; rewrite H0 in H3; try now contradict H3. 
            destruct v ; try now contradict H3.
            subst.
            apply getIndirectionProp with (StateLib.getIndexOfAddr va (CLevel (nbL - stop))) indirection;trivial.
            rename H1 into Hnotfstlevel.
            symmetry in Hnotfstlevel.
            apply levelEqBEqNatFalse0 in Hnotfstlevel.
            unfold CLevel in Hnotfstlevel.
            case_eq (lt_dec (nbL - stop) nbLevel).
           intros l Hstop.
           rewrite Hstop in Hnotfstlevel.
           simpl in *. lia.
           intros n0 Hfalse.
           rewrite Hfalse  in Hnotfstlevel.
           unfold StateLib.getNbLevel in Hnbl.
           case_eq ( gt_dec nbLevel 0);
           [intros v H | intro H] .
           rewrite H in Hnbl.
           inversion Hnbl. clear Hnbl. subst.
           simpl in *.
           contradict Hfalse.
           lia.
           intros.
           rewrite H1 in Hnbl.
           simpl in *.
           inversion Hnbl.
           symmetry. assumption.
           split. unfold not. intros Hnot.
           rewrite Hnot in H4.
           apply beq_nat_false in H4. subst. 
           now contradict H4. 
           rewrite NPeano.Nat.sub_add_distr.
           set (aux := nbL - stop) in *.
           assert (aux = CLevel aux ).
           { unfold CLevel.
             case_eq ( lt_dec aux nbLevel ).
             intros.
             simpl. intuition.
             intros. contradict H.
             unfold aux in *. cbn.
             destruct nbL.
             simpl in *. lia.       }
             
           rewrite H.
           apply levelPredMinus1 ;trivial. symmetry; assumption. 
           }
  - apply IHn in H0.
    * assumption.
    * assert ( false = levelEq l levelMin) by intuition.
      destruct H.
      clear IHn.
      clear H H0.               
      symmetry in H1. 
      apply levelPredMinus1 in H2;trivial.
      unfold CLevel in H2.
      case_eq (lt_dec (l - 1) nbLevel);
      intros; rewrite H in *.
      destruct levelpred.
      inversion H2.
      subst.
      clear H2.
      simpl.
      destruct l.
      simpl in *.
      assert (0 < nbLevel) by apply nbLevelNotZero.
      apply levelEqBEqNatFalse0 in H1.
      simpl in *.
      lia.
      destruct l.
      simpl in *.
      lia.
    * assert ( false = levelEq l levelMin) by intuition.
      destruct H.
      clear IHn.
      clear H H0.               
      symmetry in H1. 
      apply levelPredMinus1 in H2;trivial.
      unfold CLevel in H2.
      case_eq (lt_dec (l - 1) nbLevel);
      intros; rewrite H in *.
      destruct levelpred.
      inversion H2.
      subst.
      clear H2.
      simpl.
      destruct l.
      simpl in *.
      assert (0 < nbLevel) by apply nbLevelNotZero.
      apply levelEqBEqNatFalse0 in H1.
      simpl in *.
      lia.
      destruct l.
      simpl in *.
      lia. }
Qed.
           
